def _print_line_informations(choices=None, default=None):
    if choices is not None:
        print('[%s]' % '/'.join([str(chc).upper()
                                 if chc == default else str(chc)
                                 for chc in choices]), end=' ')
    if default not in {'', None}:
        if choices and default not in choices:
            for new_default, value in choices.items():
                if value == default:
                    default = new_default
                    break
        print('(default: %r)' % default, end=' ')
    return default


def ask(text, choices=None, default=None, required=False, cast=str):
    choices = [] if choices is None else [cast(chc) for chc in choices]
    while True:
        print(text, end=' ')
        default = _print_line_informations(choices, default)
        print(':', end=' ')

        result = input()

        if cast is not None:
            try:
                result = cast(result)
            except ValueError:
                print("Couldn't cast %r to %r" % (result, cast))
                continue

        if not result and required:
            print('you must provide an answer')
        elif result and choices and result not in choices:
            print('%r is not in %r' % (result, choices))
        else:
            return result


def ask_bool(text, default=True, required=False):
    default = 'yes' if default else 'no'
    result = ask(text, choices=['y', 'yes', 'n', 'no'],
                 default=default, required=required, cast=str)
    if result.lower() in {'y', 'yes'}:
        return True
    return False



import logging

TYPE_MAPPING = {'int': int, 'str': str, 'list': list, 'dict': dict,
                'bool': bool}
logger = logging.getLogger(__name__)


class NoValue:
    pass


class ConfNode:

    def __init__(self, parent=None, name='', *parameters):
        self._name = name
        self._parent = parent
        self._children = []
        self._parameters = {}
        self._load_parameters(*parameters)

    def _load_parameters(self, *parameters):
        for parameter in parameters:
            for name, value in parameter.items():
                if isinstance(value, list) and not self._has_attr(name):
                    setattr(self, name, ConfNode(self, name, *value))
                elif isinstance(value, list):
                    getattr(self, name)._load_parameters(*value)
                else:
                    self._load_parameter(name, value)
                if name not in self._children:
                    self._children.append(name)

    def _load_parameter(self, name, settings):
        if name in self._parameters:
            logger.debug('ignoring')
            return
        # something smarter that'd allow custom type
        if settings.get('type') in TYPE_MAPPING:
            settings['type'] = TYPE_MAPPING[settings['type']]
        elif isinstance(settings.get('type'), type):
            pass
        elif settings.get('type'):
            logger.warning('unknown type %r', settings['type'])
            settings['type'] = str
        else:
            settings['type'] = str
        has_default = bool(settings.get('default'))
        has_among = bool(settings.get('among'))
        settings['required'] = bool(settings.get('required'))
        settings['read_only'] = bool(settings.get('read_only'))

        path = ".".join(self._path + [name])
        if has_among:
            assert isinstance(settings['among'], list), ("parameters %s "
                    "configuration has wrong value for 'among', should be a "
                    "list, ignoring it" % path)
        if has_default and has_among:
            assert settings.get('default') in settings.get('among'), ("default"
                    " value for %r is not among the selectable values (%r)" % (
                        path, settings.get('among')))
        if has_default and settings['required']:
            raise ValueError(
                    "%r required parameter can't have default value" % path)

        if 'type' in settings and 'default' in settings:
            settings['default'] = settings['type'](settings['default'])
        self._parameters[name] = settings

    def _has_attr(self, attr):
        try:
            super().__getattribute__(attr)
            return True
        except AttributeError:
            return False

    def _set_to_path(self, path, value, overwrite=False):
        """Will set the value to the provided path. Local node if path length
        is one, a child node if path length is more that one.

        path: list
        value: the value to set
        """
        attr = path[0]
        if len(path) == 1:
            if not overwrite and self._has_attr(attr):
                return
            if 'read_only' in self._parameters[attr]:
                read_only = self._parameters[attr].pop('read_only')
                res = setattr(self, attr, value)
                self._parameters[attr]['read_only'] = read_only
                return res
            return setattr(self, attr, value)
        return getattr(self, attr)._set_to_path(path[1:], value,
                                                overwrite=overwrite)

    def __getattribute__(self, name):
        """Return a parameter of the node if this one is defined.
        Its default value if it has one.
        """
        if name.startswith('_'):
            return super().__getattribute__(name)
        if 'default' in self._parameters.get(name, {}):
            try:  # Trying to get attr, if AttributeError => is absent
                return super().__getattribute__(name)
            except AttributeError:
                return self._parameters[name]['default']
        return super().__getattribute__(name)

    def __setattr__(self, key, value):
        if key.startswith('_') or isinstance(value, ConfNode):
            return super().__setattr__(key, value)
        if key not in self._parameters:
            raise ValueError('%r is not a registered conf option' % self._path)
        if self._parameters[key].get('read_only'):
            raise AttributeError('attribute is in read only mode')
        if 'among' in self._parameters[key]:
            if value not in self._parameters[key]['among']:
                raise ValueError("%r: value %r isn't in %r" % (
                        self._path, value, self._parameters[key]['among']))
        if 'type' in self._parameters[key]:
            value = self._parameters[key]['type'](value)
        return super().__setattr__(key, value)

    @property
    def _path(self):
        if self._parent is None:
            return []
        return self._parent._path + [self._name]

    def _get_path_val_param(self):
        for name in self._children:
            if isinstance(getattr(self, name, None), ConfNode):
                yield from getattr(self, name)._get_path_val_param()
            else:
                yield self._path + [name], getattr(self, name, NoValue), \
                        self._parameters[name]



import json
import logging
from os.path import abspath, expanduser, splitext

import yaml

logger = logging.getLogger(__name__)


def read(*paths):
    any_found = False
    for path in paths:
        path = abspath(expanduser(path.strip()))
        ext = splitext(path)[1][1:]
        try:
            if ext in {'yml', 'yaml'}:
                with open(path, 'r') as fd:
                    yield path, ext, yaml.load(fd.read())
            elif ext == 'json':
                with open(path, 'r') as fd:
                    yield path, ext, json.load(fd)
            else:
                logger.error("File %r ignored: unknown type (%s)", path, ext)
                continue
            any_found = True
        except FileNotFoundError:
            logger.debug('%r not found', path)
        except PermissionError:
            logger.warning('%r: no right to read', path)
    if not any_found:
        logger.warning('no file found among %r', paths)


def _extract_value(config, path):
    if path[0] in config and len(path) == 1:
        return config[path[0]]
    elif path[0] in config:
        return _extract_value(config[path[0]], path[1:])
    else:
        raise ValueError('no %r in %r' % (path[0], config))


def extract_values(paths, config, config_file):
    for path in paths:
        try:
            yield path, _extract_value(config, path)
        except ValueError:
            logger.info('%s not found in %r', '.'.join(path), config_file)


def write(config, path):
    path = abspath(expanduser(path.strip()))
    ext = splitext(path)[1][1:]
    if ext in {'yml', 'yaml'}:
        with open(path, 'w') as fp:
            yaml.dump(config, fp)
    elif ext == 'json':
        with open(path, 'w') as fp:
            json.dump(config, fp)
    else:
        raise ValueError("couldn't make out file type, conf file path should "
                         "end with either yml, yaml or json")



from argparse import ArgumentParser

CONFIG_OPT_DEST = 'config_file_path'


def path_to_cmd_opt(path):
    return '--' + '-'.join(map(str.lower, path))


def path_to_dest(path):
    return '_'.join(path)


def get_parser(path_n_params, config_file_cmd_line):
    parser = ArgumentParser()
    parser.add_argument(*config_file_cmd_line, dest=CONFIG_OPT_DEST,
            help='set main conf file to load configuration from')
    for path, _, param in path_n_params:
        parser_kw = {}

        if param.get('no_cmd'):
            continue
        flag = param['cmd_line_opt'] if param.get('cmd_line_opt') \
                else path_to_cmd_opt(path)

        if 'type' in param:
            parser_kw['type'] = param['type']
            if param['type'] is bool and param.get('default') is False:
                param['action'], param['default'] = 'store_true', False
            elif param['type'] is bool and param.get('default') is True:
                param['action'], param['default'] = 'store_false', True
        if 'among' in param:
            parser_kw['choices'] = param['among']
        if 'help_txt' in param:
            parser_kw['help'] = param['help_txt']

        parser.add_argument(flag, dest=path_to_dest(path), **parser_kw)
    return parser


def yield_values_from_cmd(path_val_params, opts, config_file_cmd_line):
    parser = get_parser(path_val_params, config_file_cmd_line)
    cmd_line_args, _ = parser.parse_known_args(opts)
    yield getattr(cmd_line_args, CONFIG_OPT_DEST)
    for path, _, _ in path_val_params:
        value = getattr(cmd_line_args, path_to_dest(path))
        if value is not None:
            yield path, value



import os
import logging

from the_conf import files, command_line, node, interractive

logger = logging.getLogger(__name__)
DEFAULT_ORDER = 'cmd', 'files', 'env'
DEFAULT_CONFIG_FILE_CMD_LINE = '-C', '--config'
DEFAULT_CONFIG_FILE_ENVIRON = ('CONFIG_FILE',)


class TheConf(node.ConfNode):

    def __init__(self, *metaconfs, prompt_values=False,
                 cmd_line_opts=None, environ=None):
        self._source_order = list(DEFAULT_ORDER)
        self._config_files = []
        self._config_file_cmd_line = list(DEFAULT_CONFIG_FILE_CMD_LINE)
        self._config_file_environ = list(DEFAULT_CONFIG_FILE_ENVIRON)
        self._main_conf_file = None
        self._cmd_line_opts = cmd_line_opts
        self._environ = environ
        self._prompt_values = prompt_values

        def is_default(value, default):
            if not value or isinstance(value, tuple):
                return True
            return tuple(value) == default

        def set_metaconf_setting(key, metaconf, default):
            if key not in metaconf:
                return
            new_value = metaconf[key]
            if isinstance(metaconf[key], (list, tuple, set)):
                new_value = list(new_value)
            elif isinstance(new_value, (str, int, float)):
                raise TypeError('metaconf parameter %s is of unknown type %r'
                                % (key, type(new_value)))
            value = getattr(self, '_' + key)
            if is_default(value, default):
                setattr(self, '_' + key, new_value)
            else:
                value.extend(new_value)

        super().__init__()
        for mc in metaconfs:
            if isinstance(mc, str):
                _, _, mc = next(files.read(mc))
            set_metaconf_setting('source_order', mc, DEFAULT_ORDER)
            set_metaconf_setting('config_file_cmd_line',
                                 mc, DEFAULT_CONFIG_FILE_CMD_LINE)
            set_metaconf_setting('config_file_environ',
                                 mc, DEFAULT_CONFIG_FILE_ENVIRON)
            set_metaconf_setting('config_files', mc, None)

            self._load_parameters(*mc['parameters'])
        self.load()

    def _load_files(self):
        if self._config_files is None:
            return
        for conf_file, _, config in files.read(*self._config_files):
            paths = (path for path, _, _ in self._get_path_val_param())
            for path, value in files.extract_values(paths, config, conf_file):
                self._set_to_path(path, value)

    def _load_cmd(self, opts=None):
        gen = command_line.yield_values_from_cmd(
                list(self._get_path_val_param()), self._cmd_line_opts,
                self._config_file_cmd_line)
        config_file = next(gen)
        if config_file:
            self._config_files.insert(0, config_file)

        for path, value in gen:
            self._set_to_path(path, value)

    def _load_env(self, environ=None):
        if environ is None:
            environ = os.environ
        for config_env_key in self._config_file_environ:
            if config_env_key in environ:
                self._config_files.insert(0, environ[config_env_key])
        for path, _, _ in self._get_path_val_param():
            env_key = '_'.join(map(str.upper, path))
            if env_key in environ:
                self._set_to_path(path, environ[env_key])

    def load(self):
        for order in self._source_order:
            if order == 'files':
                self._load_files()
            elif order == 'cmd':
                self._load_cmd(self._cmd_line_opts)
            elif order == 'env':
                self._load_env(self._environ)
            else:
                raise Exception('unknown order %r')

        if self._prompt_values:
            self.prompt_values(False, False, False, False)

        for path, value, param in self._get_path_val_param():
            if value is node.NoValue and param.get('required'):
                raise ValueError('loading finished and %r is not set'
                        % '.'.join(path))

    def _extract_config(self):
        config = {}
        for paths, value, param in self._get_path_val_param():
            if value is node.NoValue:
                continue
            if 'default' in param and value == param['default']:
                continue
            curr_config = config
            for path in paths[:-1]:
                curr_config[path] = {}
                curr_config = curr_config[path]
            curr_config[paths[-1]] = value
        return config

    def write(self, config_file=None):
        if config_file is None and not self._config_files:
            raise ValueError('no config file to write in')

        files.write(self._extract_config(),
                    config_file or self._config_files[0])

    def prompt_values(self, only_empty=True, only_no_default=True,
            only_required=True, only_w_help=True):
        for path, value, param in self._get_path_val_param():
            if only_w_help and not param.get('help_txt'):
                continue
            if only_required and not param.get('required'):
                continue
            if only_no_default and not param.get('default'):
                continue
            if only_empty and value is not node.NoValue:
                continue
            if param.get('type') is bool:
                self._set_to_path(path, interractive.ask_bool(
                    param.get('help_txt', '.'.join(path)),
                    default=param.get('default'),
                    required=param.get('required')))
            else:
                self._set_to_path(path, interractive.ask(
                    param.get('help_txt', '.'.join(path)),
                    choices=param.get('among'), default=param.get('default'),
                    required=param.get('required'), cast=param.get('type')))



#!/usr/bin/env python3
import unittest
from setuptools import setup


def the_conf_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite


setup(name='the_conf',
      version='0.0.13',
      description='Config build from multiple sources',
      keywords='conf configuration json yaml command line environ',
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"],
      license="GPLv3",
      author="François Schmidts",
      author_email="francois@schmidts.fr",
      maintainer="François Schmidts",
      maintainer_email="francois@schmidts.fr",
      packages=['the_conf'],
      url='https://github.com/jaesivsm/the_conf/',
      install_requires=['PyYAML>=3.12'],
      test_suite='setup.the_conf_test_suite',
      )
